#ifndef RENDERER_H
#define RENDERER_H

#include <GL/freeglut.h>

#include "Volume.h"

#include <string>
#include <vector>

/**
* @class Renderer encapsulates the rendering pipeline both for the Slicing-Rendering and Volume-Rendering of the Volumes.
*/
class Renderer
{
 public:
	/**
	* @class TFPoint holds the information for one position on the Transferfunction
	*/
	struct TFPoint
	{
		unsigned int pos; //! @param pos defines the position of this RGBA quadrupple inside the Transferfunction. Ranges between 0 and the width of the Transferfunction
		unsigned char r; //! @param r defines the r (Red) value of this position inside the Transferfunction. Ranges from 0 to 255
		unsigned char g; //! @param g defines the g (Green) value of this position inside the Transferfunction. Ranges from 0 to 255
		unsigned char b; //! @param b defines the b (Blue) value of this position inside the Transferfunction. Ranges from 0 to 255
		unsigned char a; //! @param a defines the a (Alpha) value of this position inside the Transferfunction. Ranges from 0 to 255
	};

	/**
	* @class TransferFunction holds all Data which is needed to represent a Transferfunction
	*/
	struct TransferFunction
	{
		GLuint texID; //! @param texID stores the id of the texture this Transferfunction is bound when in use by OpenGL
		unsigned int width; //! @param width defines the x-size in pixel this Transferfunction has as a 1D texture
		std::vector<TFPoint> points; //! @param points holds the Points which spawn the Transferfunction and between whose will be linearly interpolated to generate the Texturedata

		unsigned char* data; //! @param data holds a pointer to the already generated and interpolated Texturedata to be used by OpenGL

		TransferFunction()
		{
			this->width = 0;
			this->texID = 0;
			this->data = 0;
			//this->points = new std::vector<TFPoint>();
		}
	};

	//! @brief one must not directly instantiate the Renderer, VolumeRendering will provide an instance
	Renderer();
	~Renderer();

	//! @brief loads a Volume from the given filename and sets it as active to be rendered
	bool loadVolume(const std::string&);

	//! @brief sets a Transferfunction to be used by the rendering
	void setTransferFunction(TransferFunction&);

	//! @brief toggles between animated or still image rendering (animated is default)
	void setAnimate(bool);
	//! @brief toggles between rendermode: 0 means Slicing-Rendering, 1 means Volume-Rendering
	void switchRenderMode(int);

	//! @brief selects the axis to be manipulated: 0 = x, 1 = y, 2 = z
	void switchAxis(int);
	//! @brief changes the axis coordinate 
	void setAxisCoord(float);

	//! @brief is being called to initalize the Renderer. MUST NOT BE CALLED, is done by RenderWindow
	void initGL();
	//! @brief is being called to to resize the Viewport. MUST NOT BE CALLED, is done by RenderWindow
	void resizeGL(int, int);
	//! @brief is being called regulary to keep up the Rendering-Loop. MUST NOT BE CALLED, is done by RenderWindow
	void paintGL();

	//! @brief is being called to process animation-steps when animation is active. MUST NOT BE CALLED, is done by RenderWindow
	void animate();

	//! @brief sets the sample distance
	void setSampleDist(float);

	//! @brief toggles between use of transferfunction or not
	void setUseTf(bool);

 private:
	//! @brief the active volume instance, 0 when nothing loaded
	Volume* volume;

	//! @brief true when animation is activated, false otherwise
	bool animationFlag;
	//! @brief 0 for Slicing-Rendering, 1 for Volume-Rendering
	int renderMode;

	//! @brief stores the rotation angles of each axes
	float rotationAngles[3];

	//! @brief the index of the axis which is about to be changed (either through animation or manually through setAxisCoord)
	int axisIndex;
	//! @brief stores the axisCoord during Slicing-Rendering between 0 and 1 (which in fact maps directly to the 3D texturecoordinate of the selected Axis: axisIndex)
	float axisCoord;
	//! @brief true when the axisCoord should be increased during animation, false when it should be decreased during animatino
	bool increaseSlice;

	//! @brief the sampledistance between each point of the volume shader
	float sampleDist;

	//! @brief a flag wheter to use a transferfunction or not 
	bool useTransferfunction;

	//! @brief the factor how long the last rendering step took
	double loopFactor;

	//! @brief width of the viewport
	GLuint width;
	//! @brief height of the viewport
	GLuint height;

	//! @brief OpenGL-ID of the Transferfunction-Texture
	GLuint transferTexture;
	//! @brief OpenGL-ID of the Volume-Texture
	GLuint volumeTexture;
	//! @brief OpenGL-ID of the Frontfaces-Texture (rendered by using FBO)
	GLuint frontTexture;
	//! @brief OpenGL-ID of the Backfaces-Texture (rendered by using FBO)
	GLuint backTexture;
	//! @brief OpenGL-ID of the Frontfaces-Rendering FBO
	GLuint frontFBO;
	//! @brief OpenGL-ID of the Backfaces-Rendering FBO
	GLuint backFBO;

	//! @brief OpenGL-ID of the Shaderprogram used for Slicing-Rendering
	GLuint slicingProgram;
	//! @brief OpenGL-ID of the Shaderprogram used for Volume-Rendering
	GLuint volumeProgram;

	GLuint slicingProgramMit;
	GLuint volumeProgramMit;

	//! @brief OpenGL-ID of the uniform location of the Volume-Texture in the Slicing-Rendering Shader
	GLint volumeSlicingTextureLocation;
	//! @brief OpenGL-ID of the uniform location of the Transferfunction-Texture in the Slicing-Rendering Shader
	GLint volumeSlicingTransferLocation;

	//! @brief OpenGL-ID of the uniform location of the Transferfunction-Texture in the Volume-Rendering Shader
	GLint transferTextureLocation;
	//! @brief OpenGL-ID of the uniform location of the Volume-Texture in the Volume-Rendering Shader
	GLint volumeTextureLocation;
	//! @brief OpenGL-ID of the uniform location of the Frontfaces-Texture in the Volume-Rendering Shader
	GLint frontTextureLocation;
	//! @brief OpenGL-ID of the uniform location of the Backfaces-Texture in the Volume-Rendering Shader
	GLint backTextureLocation;
	//! @brief OpenGL-ID of the uniform location of the Lightdirection in the Volume-Rendering Shader
	GLint lightVecLocation;
	//! @brief OpenGL-ID of the uniform location of the SampleDirection in the Volume-Rendering Shader
	GLint sampleDistLocation;

	//! @brief OpenGL-ID of the uniform location of the Volume-Texture in the MIT Slicing-Rendering Shader
	GLint volumeSlicingTextureLocationMit;
	//! @brief OpenGL-ID of the uniform location of the Transferfunction-Texture in the MIT Slicing-Rendering Shader
	GLint transferTextureLocationMit;
	
	//! @brief OpenGL-ID of the uniform location of the Volume-Texture in the MIT Volume-Rendering Shader
	GLint volumeTextureLocationMit;
	//! @brief OpenGL-ID of the uniform location of the Frontfaces-Texture in the MIT Volume-Rendering Shader
	GLint frontTextureLocationMit;
	//! @brief OpenGL-ID of the uniform location of the Backfaces-Texture in the MIT Volume-Rendering Shader
	GLint backTextureLocationMit;
	//! @brief OpenGL-ID of the uniform location of the SampleDirection in the MIT Volume-Rendering Shader
	GLint sampleDistLocationMit;

	//! @brief creates a FBO and stores the IDs of the FBO and the Texture in the given params
	void createFBO(GLuint*, GLuint*);

	//! @brief renders the Colorcube in the active bound FBO
	void renderVolumeBox();

	//! @brief renders one frame of the Slicing-Rendering
	void renderSlicing();
	//! @brief renders one frame of the Volume-Rendering
	void renderVolume();

	//! @brief calculates the animation for the Slicing-Rendering
	void animateSlicing();
	//! @brief calculates the animation for the Volume-Rendering
	void animateVolume();

	//! @brief checks if all OpenGL features are present on this hardware neede to to Volume-Rendering as implemented hiere
	void checkGLFeatures();

	//! @brief creates a Shader-Program out of two given files: the vertexprogram and the fragmentprogram
	static GLuint createProgram(const std::string&, const std::string&);
	//! @brief queries and returns uniform locations
	static GLint queryUniformLoc(GLint, const GLchar*);
	//! @brief reads text from a file to load shader sourcefiles
	static bool readShaderSource(const std::string&, std::string&);
	
};

#endif
